@@ -17,7 +17,7 @@ our %_pod_inherit_config =
 sub belongs_to {
   my ($class, $rel, $f_class, $cond, $attrs) = @_;
 
-  # assume a foreign key contraint unless defined otherwise
+  # assume a foreign key constraint unless defined otherwise
   $attrs->{is_foreign_key_constraint} = 1
     if not exists $attrs->{is_foreign_key_constraint};
   $attrs->{undef_on_null_fk} = 1
@@ -25,29 +25,39 @@ sub belongs_to {
 
   # no join condition or just a column name
   if (!ref $cond) {
+
+    my ($f_key, $guess);
+    if (defined $cond and length $cond) {
+      $f_key = $cond;
+      $guess = "caller specified foreign key '$f_key'";
+    }
+    else {
+      $f_key = $rel;
+      $guess = "using given relationship name '$rel' as foreign key column name";
+    }
+
+    $class->throw_exception(
+      "No such column '$f_key' declared yet on ${class} ($guess)"
+    )  unless $class->has_column($f_key);
+
     $class->ensure_class_loaded($f_class);
-    my %f_primaries = map { $_ => 1 } try { $f_class->_pri_cols }
-      catch {
-        $class->throw_exception( "Can't infer join condition for '$rel' on ${class}: $_");
-      };
+    my $f_rsrc = try {
+      $f_class->result_source_instance;
+    }
+    catch {
+      $class->throw_exception(
+        "Foreign class '$f_class' does not seem to be a Result class "
+      . "(or it simply did not load entirely due to a circular relation chain)"
+      );
+    };
 
-    my ($pri, $too_many) = keys %f_primaries;
-    $class->throw_exception(
-      "Can't infer join condition for '$rel' on ${class}: "
-    . "${f_class} has multiple primary keys"
-    ) if $too_many;
+    my $pri = $f_rsrc->_single_pri_col_or_die;
 
-    my $fk = defined $cond ? $cond : $rel;
-    $class->throw_exception(
-      "Can't infer join condition for '$rel' on ${class}: "
-    . "'$fk' is not a column of $class"
-    ) unless $class->has_column($fk);
-
-    $cond = { "foreign.${pri}" => "self.${fk}" };
+    $cond = { "foreign.${pri}" => "self.${f_key}" };
 
   }
   # explicit join condition
-  elsif (ref $cond) {
+  else {
     if (ref $cond eq 'HASH') { # ARRAY is also valid
       my $cond_rel;
       for (keys %$cond) {
@@ -60,13 +70,6 @@ sub belongs_to {
       $cond = $cond_rel;
     }
   }
-  # dunno
-  else {
-    $class->throw_exception(
-      'third argument for belongs_to must be undef, a column name, '.
-      'or a join condition'
-    );
-  }
 
   my $acc_type = (
     ref $cond eq 'HASH'
@@ -86,6 +89,7 @@ sub belongs_to {
   $class->add_relationship($rel, $f_class,
     $cond,
     {
+      is_depends_on => 1,
       accessor => $acc_type,
       $fk_columns ? ( fk_columns => $fk_columns ) : (),
       %{$attrs || {}}
